#!/usr/local/bin/perl

# Scan the incoming sendmail queue and find all the messages ready for
# scanning.

use strict;
package Sendmail;

my($LOCK_SH) = 1;
my($LOCK_EX) = 2;
my($LOCK_NB) = 4;
my($LOCK_UN) = 8;

# The new header I add to outgoing messages
my($XEcsHeader) = "X-ECS-Virus-Protection:";

# List all the message id's of the files in the incoming queue that are ready
# to be scanned. Will always return at least 1 message if available.
# Parameters are:
#  Sendmail incoming queue directory
#  Scanner incoming work directory
#  Ref to fill with list of definitely-clean messages
#  Ref to fill with list of possibly-dirty messages
#  #Ref to fill with number of definitely-clean messages
#  #Ref to fill with number of possibly-dirty messages
#  #Ref to fill with total size of definitely-clean messages
#  #Ref to fill with total size of possibly-dirty messages
#  Optional max number of definitely-clean messages to process
#  Optional max number of definitely-dirty messages to process
#  Optional max size of definitely-clean messages to process
#  Optional max size of possibly-dirty messages to process
sub FindMessagesToProcess {
  my($InQueueDir, $OutDir, $RClean, $RDirty,
 #    $RCleanMsgs, $RDirtyMsgs, $RCleanBytes, $RDirtyBytes,
     $MaxCleanMsgs, $MaxDirtyMsgs, $MaxCleanBytes, $MaxDirtyBytes) = @_;
  local(*QUEUE, *QFFILE);
  my($file, $RHeaders);
  my($CleanMsgs, $DirtyMsgs, $CleanBytes, $DirtyBytes);
  my($HitLimit1, $HitLimit2, $HitLimit3, $HitLimit4);

#  $$RCleanMsgs = 0;
#  $$RDirtyMsgs = 0;
#  $$RCleanBytes = 0;
#  $$RDirtyBytes = 0;
  @$RClean = ();
  @$RDirty = ();

  opendir QUEUE, $InQueueDir or return undef;
  # Keep going until end of dir or have reached every imposed limit
  while(($file = readdir QUEUE) &&
        $HitLimit1+$HitLimit2+$HitLimit3+$HitLimit4<4) {
    next unless -f "$InQueueDir/$file" && $file =~ /^qf/;
    # Try to open qf file for reading and shared-lock it.
    # Skip it if either fails.
    open QFFILE, "$InQueueDir/$file" or next;
    if (flock(QFFILE, $LOCK_SH + $LOCK_NB)) {
      $file =~ s/^qf//;
      $RHeaders = ReadQf(\*QFFILE); # Read the headers into an array of lines
      if (DefinitelyClean($RHeaders)) {
        push @$RClean, $file;
        $CleanMsgs++;
        $CleanBytes += -s "$InQueueDir/qf$file";
        $CleanBytes += -s "$InQueueDir/df$file";
        $HitLimit1 = 1 if $CleanMsgs>=$MaxCleanMsgs;
        $HitLimit2 = 1 if $CleanBytes>=$MaxCleanBytes;
      } else {
        push @$RDirty, $file;
        $DirtyMsgs++;
        $DirtyBytes += -s "$InQueueDir/qf$file";
        $DirtyBytes += -s "$InQueueDir/df$file";
        $HitLimit3 = 1 if $DirtyMsgs>=$MaxDirtyMsgs;
        $HitLimit4 = 1 if $DirtyBytes>=$MaxDirtyBytes;
        WriteHeaderFile($OutDir, $file, $RHeaders); # Write the file of headers
      }
      flock(QFFILE, $LOCK_UN);
    }
    close QFFILE;
  }
  close QUEUE;

  Log::InfoLog("Forwarding $CleanMsgs clean messages, $CleanBytes bytes");
  Log::InfoLog("Scanning $DirtyMsgs messages, $DirtyBytes bytes");
#  $$RCleanMsgs = $CleanMsgs;
#  $$RDirtyMsgs = $DirtyMsgs;
#  $$RCleanBytes = $CleanBytes;
#  $$RDirtyBytes = $DirtyBytes;
}

# Read a sendmail qf file and build an array of lines which together form
# all the mail headers.
sub ReadQf {
  my($RQf) = @_;
  my($InHeader, @results);

  $InHeader = 0;
  while(<$RQf>) {
    last if /^\./; # Bat book section 23.9.19
    $InHeader = 1 if /^H/;
    ($InHeader=0),next unless /^[H\s]/;
    s/^H//;
    push @results, $_;
  }
  return \@results;
}

# Write a fake email header file for the given message id.
sub WriteHeaderFile {
  my($OutDir, $id, $RHeaders) = @_;
  local(*HEADER);

  open(HEADER, ">$OutDir/$id.header")
    or Log::DieLog("Cannot create headers file $OutDir / $id.header, %m");
  flock(HEADER, $LOCK_EX + $LOCK_NB)
    or Log::DieLog("Someone else is tramping on $OutDir / $id.header, %m");
  print HEADER @$RHeaders;
  print HEADER "\n";
  flock(HEADER, $LOCK_UN);
  close HEADER;
}

# Delete messages which are present in both the incoming and outgoing queues
# from the outgoing queue. This will happen if I am interrupted half way
# through posting a message to the outgoing queue. Safer to start again.
# Passed refs to the lists of messages in both queues, and the location
# of the outgoing queue.
sub ClearOutQueue {
  my($InMsgs, $OutQueue) = @_;
  local(*OUTQ);
  my($file, $id);

  print "Removing duplicates from outgoing queue:\n";
  opendir OUTQ, $OutQueue or return undef;
  while($file = readdir QUEUE) {
    next unless -f "$OutQueue/$file";
    next unless $file =~ /^.f(.*)$/; # Is it a sendmail queue file?
    $id = $1;
    next unless grep /^$id$/, @$InMsgs;
    print "$OutQueue/$file\n";
    unlink "$OutQueue/$file";
  }
  closedir OUTQ;
}

# Deliver any old message ids.
sub DeliverIds {
  my($Ids, $entities, $InQ, $OutQ, $sendmail) = @_;
  local(*QF, *TF, *DF);
  my($id, $newheaders, $InHeader);

  foreach $id (@$Ids) {
    print "Sending $id\n";

    # Write the body first
    open(DF, ">$OutQ/df$id")
      or Log::DieLog("Cannot create df file $OutQ/df$id, %m");
    flock(DF, $LOCK_EX + $LOCK_NB)
      or Log::DieLog("Someone else is tramping on $OutQ / df$id, %m");
    $entities->{$id}->print_body(\*DF);
    flock(DF, $LOCK_UN);
    close DF;
    undef *DF;

    # Construct all the new headers
    $newheaders = $entities->{$id}->stringify_header;
    # Turn them into sendmail qf lines
    $newheaders =~ s/^\S/H$&/mg;
    $newheaders .= "H" . $XEcsHeader . " Found to be infected\n";

    # Copy qf file from incoming queue except for H lines
    open(TF, ">$OutQ/tf$id")
      or Log::DieLog("Cannot create tf file $OutQ/tf$id, %m");
    # Lock tf file...
    flock(TF, $LOCK_EX + $LOCK_NB)
      or Log::DieLog("Someone else is tramping on $OutQ / tf$id, %m");
    open(QF, "$InQ/qf$id")
      or Log::DieLog("Cannot read qf file $InQ/qf$id, %m");
    $InHeader = 0;
    while(<QF>) {
      last if /^\./; # Bat book section 23.9.19
      ($InHeader = 1),next if /^H/;
      next if /^\s/ && $InHeader;
      $InHeader = 0;
      print TF;
    }
    close QF;
    print TF $newheaders;
    print TF ".\n"
      or Log::DieLog("Failed to write headers for messages $id, %m");
    flock(TF, $LOCK_UN);
    close TF;
    undef *TF;
    rename "$OutQ/tf$id", "$OutQ/qf$id"
      or Log::DieLog("Cannot rename $OutQ/tf$id to $OutQ/qf$id, %m");

    # Delete the original copies from the incoming sendmail queue
    #
    # DISABLED FOR DEBUGGING IN DEVELOPMENT!!!
    #
    #unlink "$InQ/qf$id";
    #unlink "$InQ/df$id";
  }

  Sendmail::TellAbout($sendmail, $Ids);
}

# Kick sendmail into delivering each of the messages I'm passed.
sub TellAbout {
  my($sendmail, $Ids) = @_;
  my($id, $cmd);

  foreach $id (@$Ids) {
    system("$sendmail -qI$id");
  }
}

# Deliver the clean messages into the outgoing sendmail.
# Pass in the Incoming queue dir, outgoing queue dir, outgoing sendmail
# binary, list of message ids which are clean.
# Need to add a hash of headers to add to each one.
sub DeliverCleanIds {
  my($Ids, $entities, $InQ, $OutQ, $sendmail) = @_;
  my($id);
  local(*QF, *TF);

  foreach $id (@$Ids) {
    # Link the df file
    link "$InQ/df$id", "$OutQ/df$id";

    # Copy the qf file, adding extra X- header to show we checked it.
    open(TF, ">$OutQ/tf$id")
      or Log::DieLog("Cannot create clean tf file $OutQ/tf$id, %m");
    # Lock tf file...
    flock(TF, $LOCK_EX + $LOCK_NB)
      or Log::DieLog("Someone else is tramping on clean $OutQ / tf$id, %m");
    open(QF, "$InQ/qf$id")
      or Log::DieLog("Cannot read qf file $InQ/qf$id, %m");
    while(<QF>) {
      last if /^\./; # Bat book section 23.9.19
      print TF;
    }
    close QF;
    print TF "H" . $XEcsHeader . " Found to be clean\n";
    print TF ".\n"
      or Log::DieLog("Failed to write headers for clean message $id, %m");
    flock(TF, $LOCK_UN);
    close TF;
    undef *TF;
    rename "$OutQ/tf$id", "$OutQ/qf$id"
      or Log::DieLog("Cannot rename clean $OutQ/tf$id to $OutQ/qf$id, %m");

    # Delete the original copies from the incoming sendmail queue
    #
    # DISABLED FOR DEBUGGING IN DEVELOPMENT!!!
    #
    #unlink "$InQ/qf$id";
    #unlink "$InQ/df$id";
  }

  Sendmail::TellAbout($sendmail, $Ids);
}

# Given a ref to a list of header lines, work out if the message is
# 100% definitely clean and non-corruptible.
# For now, this means messages that either have no "Content-type:"
# header at all, or ones that say "Content-type: text/plain"
# Return true or false.
sub DefinitelyClean {
  my($RHeaders) = @_;
  my(@Headers, $Header, $IsClean);

  @Headers = grep /^content-type:\s+/i, @$RHeaders;
  return 0 if $#Headers>0; # Catch multiple content-type headers!
  return 1 unless @Headers; # No content-type header at all
  $Header = $Headers[0];
  chomp $Header;
  $Header =~ s/^Content-Type:\s+//i;
  #print "Header is now \"$Header\"\n";
  return 1 if $Header =~ /^text$/i;
  return 1 if $Header =~ /^text\/plain$/i;
  return 1 if $Header =~ /^text\/plain;/i;
  return 0;
}

## Find messages that are 100% definitely clean and non-corruptible.
## For now, this means messages that either have no "Content-type:"
## header at all, or ones that say "Content-type: text/plain"
## Passed in the working directory and ref to list of message ids.
## Return a ref to the list of definitely clean messages.
#sub FindCleanMessages {
#  my($SrcDir, $MsgList) = @_;
#  my($Msg, @Headers, $Header);
#  my(@Clean, $IsClean);
#  local(*HEADERS);
#
#  foreach $Msg (@$MsgList) {
#    # Look for the Content-Type: header
#    $IsClean = 0;
#    open(HEADERS, "$SrcDir/$Msg.header") or next;
#    @Headers = grep /^content-type:\s+/i, <HEADERS>;
#    close HEADERS;
#    #print "Found content types: " . join(', ', @Headers) . "\n";
#    next if $#Headers>0; # Catch multiple content-type headers!
#    $IsClean = 1 unless @Headers;
#    $Header = $Headers[0];
#    chomp $Header;
#    $Header =~ s/^Content-Type:\s+//i;
#    #print "Header is now \"$Header\"\n";
#    $IsClean = 1 if $Header =~ /^text$/i;
#    $IsClean = 1 if $Header =~ /^text\/plain$/i;
#    $IsClean = 1 if $Header =~ /^text\/plain;/i;
#    push @Clean, $Msg if $IsClean;
#  }
#
#  return \@Clean;
#}

## Write a fake email header file for each of the messages given.
#sub WriteHeaderFiles {
#  my($InDir, $OutDir, @Ids) = @_;
#
#  my($id, $line, $InHeader);
#  local(*QF, *HEADER);
#
#  foreach $id (@Ids) {
#    open(QF, "$InDir/qf$id") or next;
#    if (flock(QF, $LOCK_SH + $LOCK_NB)) {
#      open(HEADER, ">$OutDir/$id.header")
#        or Log::DieLog("Cannot create headers file $OutDir / $id.header, %m");
#      flock(HEADER, $LOCK_EX + $LOCK_NB)
#        or Log::DieLog("Someone else is tramping on $OutDir / $id.header, %m");
#      $InHeader = 0;
#      while(<QF>) {
#        last if /^\./; # Bat book section 23.9.19
#        $InHeader = 1 if /^H/;
#        ($InHeader=0),next unless /^[H\s]/;
#        s/^H//;
#        print HEADER;
#      }
#      print HEADER "\n"; # Need a blank line on the end
#      flock(HEADER, $LOCK_UN);
#      close HEADER;
#      flock(QF, $LOCK_UN);
#      close QF;
#    }
#  }
#}

1;
